open Ast;;
open Big_int;;
open Num;;

exception NotFormula of typ;;

(* a1*x1 + ... + an*xn + a0 *)
type (*new_*)linear_arith = big_int VarMap.t * big_int;;
(*
(* a1*x1 + ... + an*xn + a0 *)
type linear_arith = int VarMap.t * int;;
*)

type atomic =
| FEquality of linear_arith        (* a1*x1 + ... + an*xn + a0 == 0 *)
| FInequality of linear_arith      (* a1*x1 + ... + an*xn + a0 >= 0 *)
| FStride of linear_arith * int    (* a1*x1 + ... + an*xn + a0 divisible by step *)
;;

type formula =
| FAtomic of atomic
(* | FUnknown *)
(* | FVar of var *)
| FAnd of formula list
| FOr of formula list
| FNot of formula
| FImplies of formula * formula
| FForall of var * formula
| FExists of var * formula
;;

type (*new_*)clinear_arith = (string * string) list * string;;
(*
type clinear_arith = (int * string) list * int;;
*)

type catomic =
  CEquality of clinear_arith        (* a1*x1 + ... + an*xn + a0 == 0 *)
| CInequality of clinear_arith      (* a1*x1 + ... + an*xn + a0 >= 0 *)
| CStride of clinear_arith * int    (* a1*x1 + ... + an*xn + a0 divisible by step *)
;;

type cformula =
| CAtomic of catomic
| CAnd of cformula list
| COr of cformula list
| CNot of cformula
| CImplies of cformula * cformula
| CForall of var * cformula
| CExists of var * cformula
;;

let (*new_*)convert_linear_arith ((vars, i): (*new_*)linear_arith): (*new_*)clinear_arith =
        let varlist:(string * string) list =
                (VarMap.fold (fun k d a-> (string_of_big_int d, (string_of_var k))::a) vars [])
        in
        (varlist, string_of_big_int i)
;;

(*
let convert_linear_arith ((vars, i): linear_arith): clinear_arith =
        let varlist:(int * string) list =
                (VarMap.fold (fun k d a-> (d, (string_of_var k))::a) vars [])
        in
        (varlist, i)
;;
*)

let convert_atomic (a:atomic):catomic =
  match a with
  | FEquality la -> CEquality (convert_linear_arith la);
  | FInequality la -> CInequality (convert_linear_arith la);
  | FStride (la, step) -> CStride ((convert_linear_arith la), step)
;;


let rec convert_formula (f:formula):cformula =
  match f with
  | FAtomic a -> CAtomic (convert_atomic a)
  | FAnd flist ->
      CAnd (List.map convert_formula flist)
  | FOr flist ->
      COr (List.map convert_formula flist)
  | FNot formula ->
      CNot (convert_formula formula)
  | FImplies (p, q) ->
      CImplies ((convert_formula p), (convert_formula q))
  | FForall (var, formula) ->
      CForall (var, (convert_formula formula))
  | FExists (var, formula) ->
      CExists (var, (convert_formula formula))
;;

(*
let print_linear_arith ((vars, i):linear_arith):unit =
  VarMap.iter
    (fun x ix -> Format.printf "%d*%s+" ix (string_of_var x))
    vars;
  Format.printf "%d" i
;;
*)

let (*new_*)print_linear_arith ((vars, i):(*new_*)linear_arith):unit =
  VarMap.iter
    (fun x ix -> Format.printf "%d*%s+" (int_of_big_int ix) (string_of_var x))
    vars;
  Format.printf "%d" (int_of_big_int i)
;;

let print_atomic (prec:int) (a:atomic):unit =
  match a with
  | FEquality la -> print_linear_arith la; Format.printf "==0"
  | FInequality la -> print_linear_arith la; Format.printf ">=0"
  | FStride (la, step) -> Format.printf "%d | " step; print_linear_arith la
;;

let rec print_formula (prec:int) (f:formula):unit =
  let block = format_block prec in
  match f with
  | FAtomic a -> print_atomic prec a
  | FAnd [] -> Format.printf "true"
  | FAnd [f1] -> print_formula prec f1
  | FAnd (fh::ft) ->
      block 3 (fun () ->
        print_formula 3 fh;
        Format.printf " &&@ ";
        print_formula 3 (FAnd ft))
  | FOr [] -> Format.printf "true"
  | FOr [f1] -> print_formula prec f1
  | FOr (fh::ft) ->
      block 2 (fun () ->
        print_formula 2 fh;
        Format.printf " ||@ ";
        print_formula 2 (FOr ft))
  | FNot f ->
      block 4 (fun () ->
        Format.printf "!";
        print_formula 4 f)
  | FImplies (f1, f2) ->
      block 1 (fun () ->
        print_formula 2 f1;
        Format.printf " ==>@ ";
        print_formula 2 f2)
  | FForall (x, f1) ->
      block 0 (fun () ->
        Format.printf "all %s." (string_of_var x);
        print_formula 0 f1)
  | FExists (x, f1) ->
      block 0 (fun () ->
        Format.printf "exists %s." (string_of_var x);
        print_formula 0 f1)
;;

let (*new_*)coefficient (x:var) (vars:big_int VarMap.t):big_int =
  if VarMap.mem x vars then VarMap.find x vars else zero_big_int
;;

(*
let coefficient (x:var) (vars:int VarMap.t):int =
  if VarMap.mem x vars then VarMap.find x vars else 0
;;
*)

(*
let linear_add ((vars1, i1):linear_arith) ((vars2, i2):linear_arith):linear_arith =
  let vars = VarMap.fold
    (fun x i vars -> VarMap.add x ((coefficient x vars) + i) vars)
    vars2
    vars1
  in (vars, i1 + i2)
;;
*)

let (*new_*)linear_add ((vars1, i1):(*new_*)linear_arith) ((vars2, i2):(*new_*)linear_arith):(*new_*)linear_arith =
  let vars = VarMap.fold
    (fun x i vars -> VarMap.add x (add_big_int ((*new_*)coefficient x vars) i) vars)
    vars2
    vars1
  in (vars, (add_big_int i1 i2))
;;

(*
let linear_subtract ((vars1, i1):linear_arith) ((vars2, i2):linear_arith):linear_arith =
  let vars = VarMap.fold
    (fun x2 j2 vars1 -> VarMap.add x2 ((coefficient x2 vars1) - j2) vars1)
    vars2
    vars1
  in (vars, i1 - i2)
;;
*)

let (*new_*)linear_subtract ((vars1, i1):(*new_*)linear_arith) ((vars2, i2):(*new_*)linear_arith):(*new_*)linear_arith =
  let vars = VarMap.fold
    (fun x2 j2 vars1 -> VarMap.add x2 (sub_big_int ((*new_*)coefficient x2 vars1) j2) vars1)
    vars2
    vars1
  in (vars, (sub_big_int i1  i2))
;;

let linear_eq (l1:linear_arith) (l2:linear_arith):formula =
  FAtomic (FEquality (linear_subtract l1 l2))
;;

let linear_ne (l1:linear_arith) (l2:linear_arith):formula =
  FNot (FAtomic (FEquality (linear_subtract l1 l2)))
;;

(*
let linear_gt (l1:linear_arith) (l2:linear_arith):formula =
  (* l1 > l2  <==>  l1 - l2 > 0 <==> l1 - l2 >= 1 <==> l1 - l2 - 1 >= 0 *)
  let (vars, i) = linear_subtract l1 l2 in
  FAtomic (FInequality (vars, i - 1))
;;
*)

let (*new_*)linear_gt (l1:(*new_*)linear_arith) (l2:(*new_*)linear_arith):formula =
  (* l1 > l2  <==>  l1 - l2 > 0 <==> l1 - l2 >= 1 <==> l1 - l2 - 1 >= 0 *)
  let (vars, i) = (*new_*)linear_subtract l1 l2 in
  FAtomic (FInequality (vars, sub_big_int i (big_int_of_int 1)))
;;

let linear_ge (l1:linear_arith) (l2:linear_arith):formula =
  FAtomic (FInequality (linear_subtract l1 l2))
;;

(*
let linear_lt (l1:linear_arith) (l2:linear_arith):formula =
  (* l1 < l2  <==>  l2 - l1 > 0 <==> l2 - l1 >= 1 <==> l2 - l1 - 1 >= 0 *)
  let (vars, i) = linear_subtract l2 l1 in
  FAtomic (FInequality (vars, i - 1))
;;
*)

let (*new_*)linear_lt (l1:(*new_*)linear_arith) (l2:(*new_*)linear_arith):formula =
  (* l1 < l2  <==>  l2 - l1 > 0 <==> l2 - l1 >= 1 <==> l2 - l1 - 1 >= 0 *)
  let (vars, i) = (*new_*)linear_subtract l2 l1 in
  FAtomic (FInequality (vars, sub_big_int i (big_int_of_int 1)))
;;

let linear_le (l1:linear_arith) (l2:linear_arith):formula =
  FAtomic (FInequality (linear_subtract l2 l1))
;;

let iarith_zero = (IArith (zero_big_int, VarMap.empty));;
let iarith_const i = (IArith (i, VarMap.empty));;
let iarith_nconst i = (IInfer (i, VarMap.empty, VarMap.empty));;
let iarith_var x = (IArith (zero_big_int, VarMap.add x unit_big_int VarMap.empty));;

let weaken_int_arith (t:int_arith):int_arith =
  match t with
  | IArith (i0, vars) ->
      IInfer
        ( Big_int i0,
          VarMap.map num_of_big_int vars,
          VarMap.empty)
  | IInfer _ -> t
;;

(* compute i*t *)
let rec mult_num_arith (i:num) (t:int_arith):int_arith =
  if i =/ (Int 0) then iarith_zero else
  match t with
  | IArith _ -> mult_num_arith i (weaken_int_arith t)
  | IInfer (i0, vars, ivars) ->
      IInfer (
        i */ i0,
        VarMap.map (( */ ) i) vars,
        VarMap.map (fun (ik, tk) -> (i */ ik, tk)) ivars)
;;

(* compute i*t *)
let mult_int_arith (i:big_int) (t:int_arith):int_arith =
  if eq_big_int i zero_big_int then iarith_zero else
  match t with
  | IArith (i0, vars) ->
      IArith (mult_big_int i i0, VarMap.map (mult_big_int i) vars)
  | IInfer (i0, vars, ivars) -> mult_num_arith (Big_int i) t
;;

(* compute i*x + t *)
let rec add_num_arith_var (i:num) (x:var) (t:int_arith):int_arith =
  if i =/ (Int 0) then t else
  match t with
  | IArith _ -> add_num_arith_var i x (weaken_int_arith t)
  | IInfer (i0, vars, ivars) ->
      let vars =
        if VarMap.mem x vars then
          let isum = i +/ (VarMap.find x vars) in
          if isum =/ (Int 0) then VarMap.remove x vars else
          VarMap.add x isum vars
        else VarMap.add x i vars
      in IInfer (i0, vars, ivars)
;;

(* compute i*x + t *)
let add_int_arith_var (i:big_int) (x:var) (t:int_arith):int_arith =
  if eq_big_int i zero_big_int then t else
  match t with
  | IArith (i0, vars) ->
      let vars =
        if VarMap.mem x vars then
          let isum = add_big_int i (VarMap.find x vars) in
          if eq_big_int isum zero_big_int then VarMap.remove x vars else
          VarMap.add x isum vars
        else VarMap.add x i vars
      in IArith (i0, vars)
  | IInfer (i0, vars, ivars) -> add_num_arith_var (Big_int i) x t
;;

(* compute i*(x,ti) + t *)
let rec add_num_arith_ivar (i:num) (x:var) (ti:typ_infer ref) (t:int_arith):int_arith =
  if i =/ (Int 0) then t else
  match t with
  | IArith _ -> add_num_arith_ivar i x ti (weaken_int_arith t)
  | IInfer (i0, vars, ivars) ->
      let ivars =
        if VarMap.mem x ivars then
          let isum = i +/ (fst (VarMap.find x ivars)) in
          if isum =/ (Int 0) then VarMap.remove x ivars else
          VarMap.add x (isum, ti) ivars
        else VarMap.add x (i, ti) ivars
      in IInfer (i0, vars, ivars)
;;

(* compute t1 + t2 *)
let rec add_int_arith (t1:int_arith) (t2:int_arith):int_arith =
  match (t1, t2) with
  | (IArith _, IInfer _) -> add_int_arith (weaken_int_arith t1) t2
  | (IInfer _, IArith _) -> add_int_arith t1 (weaken_int_arith t2)
  | (IArith (i1, vars1), IArith (i2, vars2)) ->
      VarMap.fold
        (fun x i t -> add_int_arith_var i x t)
        vars2
        (IArith (add_big_int i1 i2, vars1))
  | (IInfer (i1, vars1, ivars1), IInfer (i2, vars2, ivars2)) ->
      let t = VarMap.fold
        (fun x i t -> add_num_arith_var i x t)
        vars2
        (IInfer (i1 +/ i2, vars1, ivars1)) in
      VarMap.fold
        (fun x (i, ti) t -> add_num_arith_ivar i x ti t)
        ivars2
        t
;;

(* compute t1 = t2 (without using any context information) *)
let rec eq_int_arith (t1:int_arith) (t2:int_arith):bool =
  let t = add_int_arith t1 (mult_int_arith (big_int_of_int (-1)) t2) in
  match t with
  | IArith (i, vars) ->
    (
      if eq_big_int i zero_big_int then
        VarMap.fold (fun _ _ _ -> false) vars true
      else false
    )
  | _ -> false
;;

