/**
 * Omega_Test
 *
 * Assumptions:
 * Omega lib can only add Atomic of EQ and IEQ when under AND
 * So we add and if it does not already exist.
 *
 * @author Edward Wei, Heng Huang
 */


#include <iostream>
#include <omega.h>

#include <stlport/hash_map>

using std::hash_map;

// This is pretty much an unavoidable hack
// because ml conflicts with omega
namespace ml {
extern "C" {
#include <mlvalues.h>
#include <memory.h>
}
};

using ml::local_roots;
using ml::caml__roots_block;
using ml::mlsize_t;
using ml::header_t;

#undef DEBUG

// TEMPORARY HACKS
#define Field(x, i) (((ml::value*)(x))[i])
#define Tag_val(val) (((unsigned char *) (val)) [-sizeof(ml::value)])

void formula_construct(Relation *R, F_Forall *F_forall, Formula *F, ml::value formula);

enum formula { Atomic, And, Or, Not, Implies, Forall, Exists };
enum atomic { Equality, Inequality, Stride };

struct eqstr
{
	  bool operator()(const char* s1, const char* s2) const
	  {
			return strcmp(s1, s2) == 0;
	  }
};
static hash_map<const char*, Variable_ID, std::hash<const char*>, eqstr> *variableHashMap;


/**
 * stub function called by ml.
 *
 * formula is concrete type w/ tag of constructor val.
 * field 0 is ml::value of constructor argument.
 */
extern "C" ml::value presburger(ml::value formula) {

	CAMLparam1(formula);

	/*Relation S(1);
	F_Forall* F_forall = S.add_forall();
	F_forall->declare("N");
	F_forall->declare("K");
	S.print();*/
	/*F_And *F_and = S.add_and();
	EQ_Handle eq = F_and->add_EQ();
	eq.update_const(1);

	if (S.is_satisfiable())
		cout << "S is Satisfiable";*/

	Relation R(1);
	F_Forall *F_forall = R.add_forall();

	// should add variables somewhere here
	
	assert(R.is_set());

	variableHashMap = new hash_map<const char*, Variable_ID, std::hash<const char*>, eqstr>();

#ifdef DEBUG
	cout << endl;
	cout << "Constructing Omega Relation... ";
#endif

	formula_construct(&R, F_forall, F_forall, formula);

	R.finalize();

//	cout << "\n\nEqu:";
//	R.print();

	// TODO: free up the variableHashMap
// CKH: you mean, like this?
delete variableHashMap;

	// ERIC: I changed this from R.is_satisfiable because if it only one
	// bound is satisfiable, rather then evaluating to false it will assert
	if (R.is_upper_bound_satisfiable() && R.is_lower_bound_satisfiable()) {
#ifdef DEBUG
		cout << "==> satisfiable\n\n";
#endif
		CAMLreturn (Val_int(1));
	} else {
#ifdef DEBUG
		cout << "==> unsatisfiable\n\n";
#endif
		CAMLreturn (Val_int(0));
	}
}


void linear_arith_construct(Relation *R, F_Forall *F_forall, Constraint_Handle* C, ml::value linear_arith) {

	// TODO: assert size
	/*cout << "\n Linear Arith" << linear_arith 
		<< " size " << Wosize_val(linear_arith);*/

	// the clinear_arith will be a list
	ml::value variableNode = Field(linear_arith, 0);
	
	while (variableNode != 1) {

		// data is a tuple var * int
		ml::value data = Field(variableNode, 0);
		// size should be 2
		
		coef_t coeff = String_val(Field(data, 0));
		char* variable = String_val(Field(data, 1));

#ifdef DEBUG
		cout << coeff << "* x" << variable << " + ";
#endif

		// TODO: don't know where I can free this up

		if (variableHashMap->count(variable) == 0) {

#ifdef DEBUG
			cout << " New Free_Var " << variable << "\n";
#endif

			Variable_ID globalVar = F_forall->declare(variable);
			//Free_Var_Decl *globalVar = new Free_Var_Decl(variable);
			(*variableHashMap)[variable] = globalVar;
		}

		//Variable_ID R_globalVar = R->get_local((*variableHashMap)[variable]);
		C->update_coef((*variableHashMap)[variable], coeff);

		variableNode = Field(variableNode, 1);
	}
	
	coef_t constant = String_val(Field(linear_arith, 1));
#ifdef DEBUG
	cout << constant;
#endif

	C->update_const(constant);
}


void atomic_construct(Relation *R, F_Forall *F_forall, F_And *F, ml::value atomic) {

	// TODO: assert size
	int tag = Tag_val(atomic);
	/*cout << "\n Atomic " << atomic
		<< " size " << Wosize_val(atomic);*/

	switch (tag) {
	case Equality: {

#ifdef DEBUG
		cout << "\t%%Adding eq>";
#endif
		
		EQ_Handle F_eq = F->add_EQ();

		ml::value linear_arith = Field(atomic, 0);
		linear_arith_construct(R, F_forall, &F_eq, linear_arith);

#ifdef DEBUG
		cout << " == 0\n";
#endif
		break;
	}
	case Inequality: {
		// something >= 0
		//
#ifdef DEBUG
		cout << "\t%%Adding ineq>";
#endif

		// quirky syntax different from Formula->add
		GEQ_Handle F_ieq = F->add_GEQ();
		
		ml::value linear_arith = Field(atomic, 0);
		linear_arith_construct(R, F_forall, &F_ieq, linear_arith);

#ifdef DEBUG
		cout << " >= 0\n";
#endif
		break;
	}
	case Stride:

#ifdef DEBUG
		cout << "\tStride Atomic";
#endif
		break;
	default:

#ifdef DEBUG
		cout << "\tUnkown Atomic\n";
#endif
		return;
	}
}


void formula_construct(Relation *R, F_Forall *F_forall, Formula *F, ml::value formula) {

	int constructor_tag = Tag_val(formula);

	//ml::value arguments = Code_val(arguments);
	/*cout << "\n\nformula: " << formula
		<< ":" << Wosize_val(formula) 
		<< "\nConstructor Tag: " << constructor_tag;*/
	//	<< "\nPtrArg Field " << ptr_arg;
	/*	<< "@" << Tag_val(ptr_arg);*/
		//<< ":" << Wosize_val(ptr_arg);
		
	switch(constructor_tag) {

		case Atomic: {

			// only F_And's may have Atomics
			F_And* F_and = dynamic_cast<F_And*>(F);
			if (F_and == NULL)
				F_and = F->add_and();

			atomic_construct(R, F_forall, F_and, Field(formula,0));
			break;
		}
		case Implies: {
			// P => Q  is equivalent to (not P) or Q
		
#ifdef DEBUG
			cout << "\nSee Implies ";
#endif
			ml::value formula_P = Field(formula,0);
			int tag_P = Tag_val(formula_P);
			ml::value formula_Q = Field(formula,1);
			int tag_Q = Tag_val(formula_Q);

			/*cout << "forumula_P" << formula_P 
				<< "@" << tag_P
				<< ":" << Wosize_val(formula_P) << endl;
			cout << "forumula_Q" << formula_Q 
				<< "@" << tag_Q
				<< ":" << Wosize_val(formula_Q) << endl;*/

			F_Or *R_imp = F->add_or();

#ifdef DEBUG
			cout << "%%Adding Or>";
#endif

			F_Not *R_P = R_imp->add_not();

#ifdef DEBUG
			cout << "[[%%Adding Not>";
			cout << "\nDoing P!";
#endif

			formula_construct(R, F_forall, R_P, formula_P);

#ifdef DEBUG
			cout << "\tConstructed P>]]\t";
			cout << "\n[[Doing Q!\n";
#endif

			formula_construct(R, F_forall, R_imp, formula_Q);

#ifdef DEBUG
			cout << "\tConstructed Q>]]\t";
#endif
			break;
		}
		case And: {

			/*if (argument_size != 1)
				cerr << "ERROR: Malformed arguments to constructor "
					 << argument_size << endl;
			cout << "Argument contains: " << arguments << ":"
				<< argument_size << endl;*/

			F_And *F_and = F->add_and();

#ifdef DEBUG
			cout << "%%Adding And>";
#endif
			
			// " sees a list: ";
			ml::value listnode = Field(formula,0);

			while (listnode != 1) {
				// cout << "\n" << listnode; // << ":" << Wosize_val(listnode);
				ml::value data = Field(listnode, 0);
				// cout << " data: " << data;
				formula_construct(R, F_forall, F_and, data);
				listnode = Field(listnode, 1);
			}
			break;
		}
		case Or: {

			F_Or *F_or = F->add_or();

#ifdef DEBUG
			cout << "%%Adding Or>";
#endif

			// " Or takes a list "
			ml::value listnode = Field(formula,0);

			while (listnode != 1) {
				// cout << "\n" << listnode; // << ":" << Wosize_val(listnode);
				ml::value data = Field(listnode, 0);
				// cout << " data: " << data;
				formula_construct(R, F_forall, F_or, data);
				listnode = Field(listnode, 1);
			}
			break;
		}
		case Not: {

			F_Not *F_not = F->add_not();

			// formula Size should be 1

			ml::value notform = Field(formula, 0);
			formula_construct(R, F_forall, F_not, notform);
			
			break;
		}
		case Forall: {

#ifdef DEBUG
			cout << "See Forall\n";
#endif
			break;
		}
		case Exists: {

#ifdef DEBUG
			cout << "See Exists\n";
#endif
			break;
		}
		default:

#ifdef DEBUG
			cout << "See Unkown\n";
#endif
			return;
	}
}
